'''
Offline utility tests
'''
import sys
import redis
from os import path
from mock import patch
from mock import MagicMock
from unittest import TestCase

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.stats_collector import (AbstractCounter, ThreadedCounter)


class TestStatsAbstract(TestCase):

    def test_default_key(self):
        ac = AbstractCounter()
        self.assertEqual('default_counter', ac.get_key())

    def test_overloaded_key(self):
        ac = AbstractCounter('aKey')
        self.assertEqual('aKey', ac.get_key())

    def test_not_implemented(self):
        ac = AbstractCounter()

        try:
            ac.increment()
            self.fail("increment should be abstract")
        except NotImplementedError:
            pass

        try:
            ac.value()
            self.fail("value should be abstract")
        except NotImplementedError:
            pass

        try:
            ac.expire()
            self.fail("expire should be abstract")
        except NotImplementedError:
            pass

        try:
            ac.increment()
            self.fail("increment should be abstract")
        except NotImplementedError:
            pass


class TestThreadedCounter(TestCase):
    """
    ThreadedCounter Tests
    """

    @patch.object(ThreadedCounter, '_threaded_start', side_effect=lambda: None)
    def test_cleanup_thread_resilience(self,  *args, **kwargs):
        # test cleanup thread resilience to Redis Exceptions

        redis_conn = MagicMock()
        tc = ThreadedCounter(start_time=1442671176, cycle_time=1, keep_max=1)  # Initialize the counter
        tc.setup(redis_conn=redis_conn)  # The thread ready but not started since _threaded_start is mock patched

        # Check that a non redis exception kills the thread (assurance that the correct method is tested)
        tc.expire = MagicMock(side_effect=Exception("1"))
        try:
            tc._do_thread_work()
            # this should NOT be reached
            self.assertTrue(False)
        except Exception as e:
            # this should be reached
            self.assertEqual(str(e), "1")

        # Check that a redis exception doesn't kill the thread
        tc.expire = MagicMock(side_effect=redis.RedisError())
        try:
            tc._do_thread_work()
            # this should be reached
            self.assertTrue(True)
        except:
            # this should NOT be reached
            self.assertTrue(False)





import sys
import os
import errno
import copy
import logging
import datetime
from builtins import object
from collections import OrderedDict

from pythonjsonlogger import jsonlogger
from cloghandler import ConcurrentRotatingFileHandler


class LogFactory(object):
    '''
    Goal is to manage Simple LogObject instances
    Like a Singleton
    '''
    _instance = None

    @classmethod
    def get_instance(self, **kwargs):
        if self._instance is None:
            self._instance = LogObject(**kwargs)

        return self._instance


class LogCallbackHandler:

    def __init__(self, logger):
        self.logger = logger
        self.callbacks = OrderedDict([
            ("DEBUG", []),
            ("INFO", []),
            ("WARNING", []),
            ("ERROR", []),
            ("CRITICAL", []),
        ])

    def parse_log_level(self, log_level):
        MIN_LOG_LEVEL = self.logger.level_dict['DEBUG']
        MAX_LOG_LEVEL = self.logger.level_dict['CRITICAL']

        chrs = log_level[:2]
        if chrs == '<=':
            log_level = log_level[2:]
            log_level_n = self.logger.level_dict[log_level]
            r = range(MIN_LOG_LEVEL, log_level_n + 1)
        elif chrs.startswith('<'):
            log_level = log_level[1:]
            log_level_n = self.logger.level_dict[log_level]
            r = range(MIN_LOG_LEVEL, log_level_n)
        elif chrs == '>=':
            log_level = log_level[2:]
            log_level_n = self.logger.level_dict[log_level]
            r = range(log_level_n, MAX_LOG_LEVEL + 1)
        elif chrs.startswith('>'):
            log_level = log_level[1:]
            log_level_n = self.logger.level_dict[log_level]
            r = range(log_level_n + 1, MAX_LOG_LEVEL + 1)
        elif chrs.startswith('='):
            log_level = log_level[1:]
            log_level_n = self.logger.level_dict[log_level]
            r = range(log_level_n, log_level_n + 1)
        elif chrs == '*':
            r = range(MIN_LOG_LEVEL, MAX_LOG_LEVEL + 1)
        else:
            log_level_n = self.logger.level_dict[log_level]
            r = range(log_level_n, log_level_n + 1)

        return list(r)

    def is_subdict(self, a, b):
        '''
        Return True if a is a subdict of b
        '''
        return all((k in b and b[k] == v) for k, v in a.items())

    def register_callback(self, log_level, fn, criteria=None):
        criteria = criteria or {}

        num_to_level_map = {v: k for k, v in self.logger.level_dict.items()}

        # Force normalization of WARN
        num_to_level_map[2] = "WARNING"

        log_range = self.parse_log_level(log_level)

        for log_n in log_range:
            level = num_to_level_map[log_n]
            self.callbacks[level].append((fn, criteria))

    def fire_callbacks(self, log_level, log_message=None, log_extra=None):
        log_extra = log_extra or {}

        callbacks = self.callbacks[log_level]
        for cb, criteria in callbacks:
            unmatched_criteria = criteria and not self.is_subdict(criteria, log_extra)
            if unmatched_criteria:
                continue
            else:
                cb(log_message, log_extra)


class LogObject(object):
    '''
    Easy wrapper for writing json logs to a rotating file log
    '''

    level_dict = {
        "DEBUG": 0,
        "INFO": 1,
        "WARN": 2,
        "WARNING": 2,
        "ERROR": 3,
        "CRITICAL": 4,
    }

    def __init__(self, json=False, stdout=True, name='scrapy-cluster',
                 dir='logs', file='main.log', bytes=25000000, backups=5,
                 level='INFO',
                 format='%(asctime)s [%(name)s] %(levelname)s: %(message)s',
                 propagate=False, include_extra=False):
        '''
        @param stdout: Flag to write logs to stdout or file
        @param json: Flag to write json logs with objects or just the messages
        @param name: The logger name
        @param dir: The directory to write logs into
        @param file: The file name
        @param bytes: The max file size in bytes
        @param backups: The number of backups to keep of the file
        @param level: The logging level string
        @param format: The log format
        @param propagate: Allow the log to propagate to other ancestor loggers
        @param include_extra: When not logging json, include the 'extra' param

        '''
        # set up logger
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = propagate
        self.json = json
        self.log_level = level
        self.format_string = format
        self.include_extra = include_extra
        self.cb_handler = LogCallbackHandler(self)

        if stdout:
            # set up to std out
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setLevel(logging.DEBUG)
            formatter = self._get_formatter(json)
            stream_handler.setFormatter(formatter)
            self.logger.addHandler(stream_handler)
            self._check_log_level(level)
            self.debug("Logging to stdout")
        else:
            # set up to file
            try:
                # try to make dir
                os.makedirs(dir)
            except OSError as exception:
                if exception.errno != errno.EEXIST:
                    raise

            file_handler = ConcurrentRotatingFileHandler(dir + '/' + file,
                                                         maxBytes=bytes,
                                                         backupCount=backups)
            file_handler.setLevel(logging.DEBUG)
            formatter = self._get_formatter(json)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)
            self._check_log_level(level)
            self.debug("Logging to file: {file}".format(
                file=dir + '/' + file))

    def _check_log_level(self, level):
        '''
        Ensures a valid log level

        @param level: the asked for level
        '''
        if level not in list(self.level_dict.keys()):
            self.log_level = 'DEBUG'
            self.logger.warn("Unknown log level '{lev}', defaulting to DEBUG"
                             .format(lev=level))

    def _get_formatter(self, json):
        '''
        Return the proper log formatter

        @param json: Boolean value
        '''
        if json:
            return jsonlogger.JsonFormatter()
        else:
            return logging.Formatter(self.format_string)

    def debug(self, message, extra={}):
        '''
        Writes an error message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        if self.level_dict['DEBUG'] >= self.level_dict[self.log_level]:
            extras = self.add_extras(extra, "DEBUG")
            self._write_message(message, extras)
            self.cb_handler.fire_callbacks('DEBUG', message, extras)

    def info(self, message, extra={}):
        '''
        Writes an info message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        if self.level_dict['INFO'] >= self.level_dict[self.log_level]:
            extras = self.add_extras(extra, "INFO")
            self._write_message(message, extras)
            self.cb_handler.fire_callbacks('INFO', message, extras)

    def warn(self, message, extra={}):
        '''
        Writes a warning message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        self.warning(message, extra)

    def warning(self, message, extra={}):
        '''
        Writes a warning message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        if self.level_dict['WARNING'] >= self.level_dict[self.log_level]:
            extras = self.add_extras(extra, "WARNING")
            self._write_message(message, extras)
            self.cb_handler.fire_callbacks('WARNING', message, extras)

    def error(self, message, extra={}):
        '''
        Writes an error message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        if self.level_dict['ERROR'] >= self.level_dict[self.log_level]:
            extras = self.add_extras(extra, "ERROR")
            self._write_message(message, extras)
            self.cb_handler.fire_callbacks('ERROR', message, extras)

    def critical(self, message, extra={}):
        '''
        Writes a critical message to the log

        @param message: The message to write
        @param extra: The extras object to pass in
        '''
        if self.level_dict['CRITICAL'] >= self.level_dict[self.log_level]:
            extras = self.add_extras(extra, "CRITICAL")
            self._write_message(message, extras)
            self.cb_handler.fire_callbacks('CRITICAL', message, extras)

    def _write_message(self, message, extra):
        '''
        Writes the log output
        @param message: The message to write
        @param extra: The potential object to write
        '''
        if not self.json:
            self._write_standard(message, extra)
        else:
            self._write_json(message, extra)

    def _write_standard(self, message, extra):
        '''
        Writes a standard log statement

        @param message: The message to write
        @param extra: The object to pull defaults from
        '''
        level = extra['level']
        if self.include_extra:
            del extra['timestamp']
            del extra['level']
            del extra['logger']
            if len(extra) > 0:
                message += " " + str(extra)

        if level == 'INFO':
            self.logger.info(message)
        elif level == 'DEBUG':
            self.logger.debug(message)
        elif level == 'WARNING':
            self.logger.warning(message)
        elif level == 'ERROR':
            self.logger.error(message)
        elif level == 'CRITICAL':
            self.logger.critical(message)
        else:
            self.logger.debug(message)

    def _write_json(self, message, extra):
        '''
        The JSON logger doesn't obey log levels

        @param message: The message to write
        @param extra: The object to write
        '''
        self.logger.info(message, extra=extra)

    @property
    def name(self):
        '''
        Returns the logger name
        '''
        return self.logger.name

    def add_extras(self, dict, level):
        '''
        Adds the log level to the dict object
        '''
        my_copy = copy.deepcopy(dict)
        if 'level' not in my_copy:
            my_copy['level'] = level
        if 'timestamp' not in my_copy:
            my_copy['timestamp'] = self._get_time()
        if 'logger' not in my_copy:
            my_copy['logger'] = self.name
        return my_copy

    def _get_time(self):
        '''
        Returns the system time
        '''
        return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def register_callback(self, log_level, fn, criteria=None):
        '''
        Passthrough for registering callbacks to preserve backwards
        compatibility
        '''
        return self.cb_handler.register_callback(log_level, fn, criteria)



import imp
import logging
import importlib
from builtins import object

log = logging.getLogger(__name__)


class SettingsWrapper(object):
    '''
    Wrapper for loading settings files and merging them with overrides
    '''

    my_settings = {}
    ignore = [
        '__builtins__',
        '__file__',
        '__package__',
        '__doc__',
        '__name__',
        '__spec__',
        '__loader__',
        '__cached__',
    ]

    def _init__(self):
        pass

    def load(self, local='localsettings.py', default='settings.py'):
        '''
        Load the settings dict

        @param local: The local settings filename to use
        @param default: The default settings module to read
        @return: A dict of the loaded settings
        '''
        self._load_defaults(default)
        self._load_custom(local)

        return self.settings()

    def load_from_string(self, settings_string='', module_name='customsettings'):
        '''
        Loads settings from a settings_string. Expects an escaped string like
        the following:
            "NAME=\'stuff\'\nTYPE=[\'item\']\n"

        @param settings_string: The string with your settings
        @return: A dict of loaded settings
        '''
        try:
            mod = imp.new_module(module_name)
            exec(settings_string, mod.__dict__)
        except TypeError:
            log.warning("Could not import settings")
        self.my_settings = {}
        try:
            self.my_settings = self._convert_to_dict(mod)
        except ImportError:
            log.warning("Settings unable to be loaded")

        return self.settings()

    def settings(self):
        '''
        Returns the current settings dictionary
        '''
        return self.my_settings

    def _load_defaults(self, default='settings.py'):
        '''
        Load the default settings
        '''
        if default[-3:] == '.py':
            default = default[:-3]

        self.my_settings = {}
        try:
            settings = importlib.import_module(default)
            self.my_settings = self._convert_to_dict(settings)
        except ImportError:
            log.warning("No default settings found")

    def _load_custom(self, settings_name='localsettings.py'):
        '''
        Load the user defined settings, overriding the defaults

        '''
        if settings_name[-3:] == '.py':
            settings_name = settings_name[:-3]

        new_settings = {}
        try:
            settings = importlib.import_module(settings_name)
            new_settings = self._convert_to_dict(settings)
        except ImportError:
            log.info("No override settings found")

        for key in new_settings:
            if key in self.my_settings:
                item = new_settings[key]
                if isinstance(item, dict) and \
                        isinstance(self.my_settings[key], dict):
                    for key2 in item:
                        self.my_settings[key][key2] = item[key2]
                else:
                    self.my_settings[key] = item
            else:
                self.my_settings[key] = new_settings[key]

    def _convert_to_dict(self, setting):
        '''
        Converts a settings file into a dictionary, ignoring python defaults

        @param setting: A loaded setting module
        '''
        the_dict = {}
        set = dir(setting)
        for key in set:
            if key in self.ignore:
                continue
            value = getattr(setting, key)
            the_dict[key] = value

        return the_dict



from __future__ import print_function
from future import standard_library

import datetime
import unittest
from builtins import next
from unittest import TestCase
standard_library.install_aliases()

import sys
from os import path

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

import json
import redis
import threading, time
from kafka import KafkaConsumer
from twisted.internet import reactor
from scrapy.crawler import CrawlerRunner
from redis.exceptions import ConnectionError
from bhfutils.spiders.link_spider import LinkSpider
from scrapy.utils.project import get_project_settings


class CustomSpider(LinkSpider):
    '''
    Overridden link spider for testing
    '''
    name = "test-spider"


class TestLinkSpider(TestCase):
    example_feed = "{\"allowed_domains\":null,\"allow_regex\":null,\"" \
                   "crawlid\":\"abc12345\",\"url\":\"http://books.toscrape.com/\",\"expires\":0,\"" \
                   "ts\":1461549923.7956631184,\"priority\":1,\"deny_regex\":null,\"" \
                   "cookie\":null,\"attrs\":null,\"appid\":\"test\",\"spiderid\":\"" \
                   "test-spider\",\"useragent\":null,\"deny_extensions\":null,\"maxdepth\":0, \"domain_max_pages\":0}"

    example_feed_max = "{\"allowed_domains\":[\"toscrape.com\"],\"allow_regex\":null,\"" \
                       "crawlid\":\"abc1234567\",\"url\":\"http://books.toscrape.com/\",\"expires\":0,\"" \
                       "ts\":1461549923.7956631184,\"priority\":1,\"deny_regex\":null,\"" \
                       "cookie\":null,\"attrs\":null,\"appid\":\"test\",\"spiderid\":\"" \
                       "test-spider\",\"useragent\":null,\"deny_extensions\":null,\"maxdepth\":3, \"domain_max_pages\":4}"

    def setUp(self):
        self.settings = get_project_settings()
        self.settings.set('KAFKA_TOPIC_PREFIX', "demo_test")
        # set up redis
        self.redis_conn = redis.Redis(host=self.settings['REDIS_HOST'],
                                      port=self.settings['REDIS_PORT'],
                                      db=self.settings['REDIS_DB'],
                                      password=self.settings['REDIS_PASSWORD'],
                                      decode_responses=True)
        try:
            self.redis_conn.info()
        except ConnectionError:
            print("Could not connect to Redis")
            # plugin is essential to functionality
            sys.exit(1)

        # clear out older test keys if any
        keys = self.redis_conn.keys("test-spider:*")
        for key in keys:
            self.redis_conn.delete(key)

        # set up kafka to consumer potential result
        self.consumer = KafkaConsumer(
            "demo_test.crawled_firehose",
            bootstrap_servers=self.settings['KAFKA_HOSTS'],
            group_id="demo-id",
            auto_commit_interval_ms=10,
            consumer_timeout_ms=5000,
            auto_offset_reset='earliest',
            value_deserializer=lambda m: m.decode('utf-8')
        )
        print("setUp end")
        time.sleep(1)

    def test_crawler_process(self):
        print("start test_crawler_process")
        runner = CrawlerRunner(self.settings)
        d = runner.crawl(CustomSpider)
        d.addBoth(lambda _: reactor.stop())
        # add crawl to redis
        key = "test-spider:toscrape.com:queue"
        self.redis_conn.zadd(key, {self.example_feed: -80})
        self.redis_conn.zadd(key, {self.example_feed_max: -90})

        # run the spider, give 20 seconds to see the urls and crawl them
        # and send to kafka. Then we kill the reactor
        total_time = 60

        def thread_func():
            time.sleep(total_time)
            reactor.stop()

        thread = threading.Thread(target=thread_func)
        thread.start()
        reactor.run()
        print("thread run")

        message_count = 0
        max_message_count = 0

        start_time = datetime.datetime.now()
        # give the consumer X seconds to consume all pages
        while (datetime.datetime.now() - start_time).total_seconds() < total_time:
            try:
                m = None
                m = next(self.consumer)
            except StopIteration as e:
                pass

            if m is None:
                pass
            else:
                print(m.value)
                the_dict = json.loads(m.value)
                if the_dict is not None:
                    if the_dict['appid'] == 'test' \
                            and the_dict['crawlid'] == 'abc1234567':
                        max_message_count += 1
                    elif the_dict['appid'] == 'test' \
                            and the_dict['crawlid'] == 'abc12345':
                        message_count += 1

        self.assertEqual(message_count, 1)
        self.assertEqual(max_message_count, 4)

    def tearDown(self):
        keys = self.redis_conn.keys('stats:crawler:*:test-spider:*')
        keys = keys + self.redis_conn.keys('test-spider:*')
        for key in keys:
            self.redis_conn.delete(key)

        # if for some reason the tests fail, we end up falling behind on
        # the consumer
        for m in self.consumer:
            pass
        self.consumer.close()


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import argparse


class ArgparseHelper(argparse._HelpAction):
    '''
    Used to help print top level '--help' arguments from argparse
    when used with subparsers

    Usage:
    from scutils.arparse_helper import ArgparseHelper
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-h', '--help', action=ArgparseHelper,
                        help='show this help message and exit')
    # add subparsers below these lines
    '''

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        print()

        subparsers_actions = [
            action for action in parser._actions
            if isinstance(action, argparse._SubParsersAction)]
        for subparsers_action in subparsers_actions:
            for choice, subparser in list(subparsers_action.choices.items()):
                print("Command '{}'".format(choice))
                print(subparser.format_usage())

        parser.exit()



'''
Offline utility tests
'''
import sys
from os import path
from mock import MagicMock
from unittest import TestCase
from redis.exceptions import WatchError

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.redis_throttled_queue import RedisThrottledQueue


class TestUnmoderatedRedisThrottledQueue(TestCase):

    def setUp(self):
        # limit is 2 hits in the window
        self.queue = RedisThrottledQueue(MagicMock(), MagicMock(), 1, 2)

    def test_unmoderated(self):
        # an unmoderated queue is really just testing the number
        # of hits in a given window
        self.queue.redis_conn.zcard = MagicMock(return_value=0)
        self.assertTrue(self.queue.allowed())

        self.queue.redis_conn.zcard = MagicMock(return_value=1)
        self.assertTrue(self.queue.allowed())

        self.queue.redis_conn.zcard = MagicMock(return_value=2)
        self.assertFalse(self.queue.allowed())

        # mock exception raised even with good hits
        self.queue.redis_conn.zcard = MagicMock(return_value=0,
                                                side_effect=WatchError)
        self.assertFalse(self.queue.allowed())


class TestModeratedRedisThrottledQueue(TestCase):

    def setUp(self):
        self.queue = RedisThrottledQueue(MagicMock(), MagicMock(), 4, 2, True)

    def test_moderated(self):
        # a moderated queue should pop ~ every x seconds
        # we already tested the window limit in the unmoderated test
        self.queue.is_moderated = MagicMock(return_value=True)
        self.assertFalse(self.queue.allowed())

        self.queue.is_moderated = MagicMock(return_value=False)
        self.queue.test_hits = MagicMock(return_value=True)
        self.assertTrue(self.queue.allowed())

        # mock exception raised even with good moderation
        self.queue.test_hits = MagicMock(side_effect=WatchError)
        self.assertFalse(self.queue.allowed())

class TestModeratedElasticRedisThrottledQueue(TestCase):

    def setUp(self):
        self.queue = RedisThrottledQueue(MagicMock(), MagicMock(), 4, 2, True,
                                         elastic=True)

    def test_moderated(self):
        # test elastic kick in hasnt happened yet
        self.queue.is_moderated = MagicMock(return_value=True)
        self.queue.elastic_kick_in = 0
        self.assertFalse(self.queue.allowed())

        # kick in overrides, even though we were moderated
        self.queue.elastic_kick_in = self.queue.limit
        self.queue.check_elastic = MagicMock(return_value=True)
        self.queue.test_hits = MagicMock(return_value=True)
        self.assertTrue(self.queue.allowed())



'''
Offline utility tests
'''

import os
import six
import sys
import json
import copy
from os import path
from mock import MagicMock
from unittest import TestCase
from testfixtures import LogCapture

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.log_factory import LogFactory, LogObject


class TestLogFactory(TestCase):
    def setUp(self):
        self.logger = LogFactory.get_instance(name='test',
                                              dir='./', level='DEBUG',
                                              propagate=True)

    def test_name_as_property(self):
        self.assertEqual('test', self.logger.name)

    def test_debug_log(self):
        self.logger.log_level = 'DEBUG'
        with LogCapture() as l:
            self.logger.debug('debug message')
            self.logger.info('info message')
        l.check(
            ('test', 'DEBUG', 'debug message'),
            ('test', 'INFO', 'info message'),
        )

    def test_info_log(self):
        self.logger.log_level = 'INFO'
        with LogCapture() as l:
            self.logger.debug('debug message')
            self.logger.info('info message')
            self.logger.warn('warn message')
        l.check(
            ('test', 'INFO', 'info message'),
            ('test', 'WARNING', 'warn message'),
        )

    def test_warn_log(self):
        self.logger.log_level = 'WARN'
        with LogCapture() as l:
            self.logger.info('info message')
            self.logger.warn('warn message')
            self.logger.error('error message')
        l.check(
            ('test', 'WARNING', 'warn message'),
            ('test', 'ERROR', 'error message'),
        )

    def test_error_log(self):
        self.logger.log_level = 'ERROR'
        with LogCapture() as l:
            self.logger.warn('warn message')
            self.logger.error('error message')
            self.logger.critical('critical message')

        l.check(
            ('test', 'ERROR', 'error message'),
            ('test', 'CRITICAL', 'critical message'),
        )

    def test_critical_log(self):
        self.logger.log_level = 'CRITICAL'
        with LogCapture() as l:
            self.logger.error('error message')
            self.logger.critical('critical message')

        l.check(
            ('test', 'CRITICAL', 'critical message'),
        )

    def test_include_extra_log(self):
        self.logger.log_level = 'INFO'
        self.logger.include_extra = True
        with LogCapture() as l:
            self.logger.info('info message', {"test": 1})

        l.check(
            ('test', 'INFO', "info message {'test': 1}"),
        )

        # don't output an empty dict
        with LogCapture() as l:
            self.logger.info('info message2', {})

        l.check(
            ('test', 'INFO', "info message2"),
        )

class TestLogJSONFile(TestCase):
    def setUp(self):
        self.logger = LogObject(name='test', json=True,
                                dir='.', level='INFO', stdout=False,
                                file='test.log')
        self.test_file = './test'

    def test_log_file_json(self):
        self.logger._get_time = MagicMock(return_value='2015-11-12T10:11:12.0Z')
        self.logger.info("Test log")
        with open(self.test_file + '.log', 'r') as f:
            read_data = f.read()
            the_dict = json.loads(read_data)
            six.assertCountEqual(self, the_dict, {
                "message": "Test log",
                "level": "INFO",
                "logger":"test",
                "timestamp":"2015-11-12T10:11:12.0Z"})

    def test_preserve_extra(self):
        self.logger.log_level = 'DEBUG'
        before = {"some": "dict", "a": 1}
        preserve = copy.deepcopy(before)
        with LogCapture() as l:
            self.logger.debug("my message", extra=before)

        self.assertEqual(preserve, before)

    def tearDown(self):
        os.remove(self.test_file + '.log')
        os.remove(self.test_file + '.lock')

class TestLogCallbacks(TestCase):
    def setUp(self):
        self.logger = LogObject(name='test', json=True,
                                dir='.', level='INFO', stdout=False,
                                )
        self.logger.x = 1

    def test_log_callbacks_integration(self):
        def add_1(log_message=None, log_extra=None):
            self.logger.x += 1

        def negate(log_message=None, log_extra=None):
            self.logger.x *= -1

        def multiply_5(log_message=None, log_extra=None):
            self.logger.x *= 5

        self.logger.register_callback('<=INFO', add_1, {'key': 'val1'})
        self.logger.register_callback('<=INFO', negate, {'key': 'val2'})
        self.logger.register_callback('<=INFO', multiply_5)

        self.logger.x = 1
        self.logger.log_level = 'INFO'
        self.logger.info('info message')
        self.assertEqual(5, self.logger.x)

        self.logger.x = 1
        self.logger.log_level = 'INFO'
        self.logger.info('info message', extra={'key': 'val1'})
        self.assertEqual(10, self.logger.x)

        self.logger.x = 1
        self.logger.log_level = 'INFO'
        self.logger.info('info message', extra={'key': 'val2'})
        self.assertEqual(-5, self.logger.x)

        # Callback shouldn't fire
        self.logger.x = 1
        self.logger.log_level = 'CRITICAL'
        self.logger.info('info message')
        self.assertEqual(1, self.logger.x)

        # Callback shouldn't fire
        self.logger.x = 1
        self.logger.log_level = 'INFO'
        self.logger.warning('warning message')
        self.assertEqual(1, self.logger.x)

    def test_parse_log_level(self):
        log_range = self.logger.cb_handler.parse_log_level("<=INFO")
        self.assertEqual([0,1], log_range)

        log_range = self.logger.cb_handler.parse_log_level("<INFO")
        self.assertEqual([0], log_range)

        log_range = self.logger.cb_handler.parse_log_level(">=WARNING")
        self.assertEqual([2,3,4], log_range)

        log_range = self.logger.cb_handler.parse_log_level(">WARN")
        self.assertEqual([3,4], log_range)

        log_range = self.logger.cb_handler.parse_log_level("=INFO")
        self.assertEqual([1], log_range)

        log_range = self.logger.cb_handler.parse_log_level("CRITICAL")
        self.assertEqual([4], log_range)

    def test_register_callback(self):
        def add_1(log_obj, log_message=None, log_extra=None):
            pass

        def add_2(log_obj, log_message=None, log_extra=None):
            pass

        def add_3(log_obj, log_message=None, log_extra=None):
            pass

        def add_4(log_obj, log_message=None, log_extra=None):
            pass

        self.logger.register_callback('>=INFO', add_1)
        self.logger.register_callback('<=WARN', add_2)
        self.logger.register_callback('ERROR', add_3)
        self.logger.register_callback('*', add_4)

        callbacks = [cb for cb,criteria in self.logger.cb_handler.callbacks['DEBUG']]
        self.assertEqual([add_2, add_4], callbacks)

        callbacks = [cb for cb,criteria in self.logger.cb_handler.callbacks['INFO']]
        self.assertEqual([add_1, add_2, add_4], callbacks)

        callbacks = [cb for cb,criteria in self.logger.cb_handler.callbacks['WARNING']]
        self.assertEqual([add_1, add_2, add_4], callbacks)

        callbacks = [cb for cb,criteria in self.logger.cb_handler.callbacks['ERROR']]
        self.assertEqual([add_1, add_3, add_4], callbacks)

        callbacks = [cb for cb,criteria in self.logger.cb_handler.callbacks['CRITICAL']]
        self.assertEqual([add_1, add_4], callbacks)

    def test_fire_callbacks_basic_1(self):
        def add_1(log_message=None, log_extra=None):
            self.logger.x += 1

        def negate(log_message=None, log_extra=None):
            self.logger.x *= -1

        def multiply_5(log_message=None, log_extra=None):
            self.logger.x *= 5

        self.logger.register_callback('<=INFO', add_1)
        self.logger.register_callback('INFO', negate)
        self.logger.register_callback('<CRITICAL', add_1)
        self.logger.register_callback('>DEBUG', multiply_5)

        self.logger.x = 0
        self.logger.log_level = 'DEBUG'
        self.logger.cb_handler.fire_callbacks('DEBUG')
        self.assertEqual(2, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'INFO'
        self.logger.cb_handler.fire_callbacks('INFO')
        self.assertEqual(0, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'WARNING'
        self.logger.cb_handler.fire_callbacks('WARNING')
        self.assertEqual(5, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'ERROR'
        self.logger.cb_handler.fire_callbacks('ERROR')
        self.assertEqual(5, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'CRITICAL'
        self.logger.cb_handler.fire_callbacks('CRITICAL')
        self.assertEqual(0, self.logger.x)

    def test_fire_callbacks_basic_2(self):
        def add_1(log_message=None, log_extra=None):
            self.logger.x += 1

        def negate(log_message=None, log_extra=None):
            self.logger.x *= -1

        def multiply_5(log_message=None, log_extra=None):
            self.logger.x *= 5

        self.logger.register_callback('>DEBUG', add_1)
        self.logger.register_callback('=WARNING', negate)
        self.logger.register_callback('<INFO', add_1)
        self.logger.register_callback('>=INFO', multiply_5)
        self.logger.register_callback('*', add_1)

        self.logger.x = 0
        self.logger.log_level = 'DEBUG'
        self.logger.cb_handler.fire_callbacks('DEBUG')
        self.assertEqual(2, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'INFO'
        self.logger.cb_handler.fire_callbacks('INFO')
        self.assertEqual(6, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'WARNING'
        self.logger.cb_handler.fire_callbacks('WARNING')
        self.assertEqual(-4, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'ERROR'
        self.logger.cb_handler.fire_callbacks('ERROR')
        self.assertEqual(6, self.logger.x)

        self.logger.x = 0
        self.logger.log_level = 'CRITICAL'
        self.logger.cb_handler.fire_callbacks('CRITICAL')
        self.assertEqual(6, self.logger.x)

    def test_preserve_data(self):
        self.logger._get_time = MagicMock(return_value='2015-11-12T10:11:12.0Z')
        message = "THIS IS A TEST"
        extras = {"key": "value", 'a': [1, 2, 3]}
        extras_add = self.logger.add_extras(extras, 'INFO')

        def cb(log_message=None, log_extra=None):
            self.assertEqual(log_message, message)
            self.assertEqual(log_extra, extras_add)

        self.logger.register_callback('>DEBUG', cb)
        self.logger.log_level = 'INFO'
        self.logger.info(message, extras)

    def tearDown(self):
        os.remove('main.log')
        os.remove('main.lock')



# coding=utf-8
import mock
import ujson
import base64
from scrapy import Item
from copy import deepcopy
from mock import MagicMock
from builtins import object
from unittest import TestCase
from crawling.items import RawResponseItem
from kafka.errors import KafkaTimeoutError
from crawling.pipelines import (LoggingBeforePipeline, KafkaPipeline)


class ItemMixin(object):

    def _get_item(self):
        item = RawResponseItem()
        item['appid'] = 'app'
        item['crawlid'] = 'crawlid'
        item['attrs'] = {}
        item["url"] = "http://dumb.com"
        item["response_url"] = "http://dumb.com"
        item["statusCode"] = 200
        item["status_msg"] = "OK"
        item["response_headers"] = {}
        item["request_headers"] = {}
        item["body"] = "text"
        item["links"] = []
        item["encoding"] = "utf-8"

        return item

    def _get_internationalized_utf8_item(self):
        item = RawResponseItem()
        item['appid'] = 'app'
        item['crawlid'] = 'crawlid'
        item['attrs'] = {}
        item["url"] = "http://dumb.com"
        item["response_url"] = "http://dumb.com"
        item["statusCode"] = 200
        item["status_msg"] = "OK"
        item["response_headers"] = {}
        item["request_headers"] = {}
        item["body"] = u"This is a test - Αυτό είναι ένα τεστ - 这是一个测试 - これはテストです"
        item["links"] = []
        item["encoding"] = "utf-8"

        return item

    def _get_internationalized_iso_item(self):
        item = RawResponseItem()
        item['appid'] = 'app'
        item['crawlid'] = 'crawlid'
        item['attrs'] = {}
        item["url"] = "http://dumb.com"
        item["response_url"] = "http://dumb.com"
        item["statusCode"] = 200
        item["status_msg"] = "OK"
        item["response_headers"] = {}
        item["request_headers"] = {}
        # Fill the item["body"] with the string 'αυτό είναι ένα τεστ' that was encoded in iso-8859-7
        # using iconv and further encoded in base64 in order to store it inside this file.
        item["body"] = base64.b64decode('4fX0/CDl3+3h6SDd7eEg9OXz9Ao=')
        item["links"] = []
        item["encoding"] = "iso-8859-7"

        return item


class TestLoggingBeforePipeline(TestCase, ItemMixin):

    def setUp(self):
        self.pipe = LoggingBeforePipeline(MagicMock())
        self.pipe.logger.name = "crawler"

    def test_process_item(self):
        item = self._get_item()

        spider = MagicMock()
        spider.name = "link"

        self.pipe.logger.info = MagicMock(side_effect=Exception("info"))
        try:
            self.pipe.process_item(item, spider)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "info")

        # test unknown item
        class WeirdItem(Item):
            pass
        item2 = WeirdItem()

        self.pipe.logger.warn = MagicMock(side_effect=Exception("warn"))
        try:
            self.pipe.process_item(item2, spider)
            self.assertFalse(True)
        except Exception as e:
            self.assertEqual(str(e), "warn")


class TestKafkaPipeline(TestCase, ItemMixin):

    def setUp(self):
        self.pipe = KafkaPipeline(MagicMock(), 'prefix', MagicMock(), False, False)
        self.pipe.producer.send = MagicMock()
        self.pipe._get_time = MagicMock(return_value='the time')

    @mock.patch('traceback.format_exc', return_value='traceback')
    def test_process_item(self, e):
        item = self._get_item()
        spider = MagicMock()
        spider.name = "link"

        # test normal send, no appid topics
        self.pipe.process_item(item, spider)
        expected = '{"appid":"app","attrs":{},"body":"text","crawlid":"crawlid","encoding":"utf-8","links":[],"request_headers":{},"response_headers":{},"response_url":"http:\\/\\/dumb.com","statusCode":200,"status_msg":"OK","timestamp":"the time","url":"http:\\/\\/dumb.com"}'
        self.pipe.producer.send.assert_called_once_with('prefix.crawled_firehose',
                                                        expected)
        self.pipe.producer.send.reset_mock()

        # test normal send, with appids
        item = self._get_item()
        self.pipe.appid_topics = True
        self.pipe.process_item(item, spider)
        self.pipe.producer.send.assert_called_with('prefix.crawled_app',
                                                    expected)
        self.pipe.producer.send.reset_mock()

        # test base64 encode
        item = self._get_item()
        self.pipe.appid_topics = False
        self.pipe.use_base64 = True
        self.pipe.process_item(item, spider)
        expected = '{"appid":"app","attrs":{},"body":"dGV4dA==","crawlid":"crawlid","encoding":"utf-8","links":[],"request_headers":{},"response_headers":{},"response_url":"http:\\/\\/dumb.com","statusCode":200,"status_msg":"OK","timestamp":"the time","url":"http:\\/\\/dumb.com"}'
        self.pipe.producer.send.assert_called_with('prefix.crawled_firehose',
                                                        expected)

        # test base64 encode/decode with utf-8 encoding
        item = self._get_internationalized_utf8_item()
        self.pipe.appid_topics = False
        self.pipe.use_base64 = True
        self.pipe.process_item(item, spider)
        expected = '{"appid":"app","attrs":{},"body":"VGhpcyBpcyBhIHRlc3QgLSDOkc+Fz4TPjCDOtc6vzr3Osc65IM6tzr3OsSDPhM61z4PPhCAtIOi\\/meaYr+S4gOS4qua1i+ivlSAtIOOBk+OCjOOBr+ODhuOCueODiOOBp+OBmQ==","crawlid":"crawlid","encoding":"utf-8","links":[],"request_headers":{},"response_headers":{},"response_url":"http:\\/\\/dumb.com","statusCode":200,"status_msg":"OK","timestamp":"the time","url":"http:\\/\\/dumb.com"}'
        self.pipe.producer.send.assert_called_with('prefix.crawled_firehose',
                                                        expected)
        # unpack the arguments used for the previous assertion call
        call_args, call_kwargs = self.pipe.producer.send.call_args
        crawl_args_dict = ujson.loads(call_args[1])
        decoded_string = base64.b64decode(crawl_args_dict['body']).decode(crawl_args_dict['encoding'])
        self.assertEquals(decoded_string, item.get('body'))

        # test base64 encode/decode with iso encoding
        item = self._get_internationalized_iso_item()
        self.pipe.appid_topics = False
        self.pipe.use_base64 = True
        self.pipe.process_item(item, spider)
        expected = '{"appid":"app","attrs":{},"body":"4fX0\\/CDl3+3h6SDd7eEg9OXz9Ao=","crawlid":"crawlid","encoding":"iso-8859-7","links":[],"request_headers":{},"response_headers":{},"response_url":"http:\\/\\/dumb.com","statusCode":200,"status_msg":"OK","timestamp":"the time","url":"http:\\/\\/dumb.com"}'
        self.pipe.producer.send.assert_called_with('prefix.crawled_firehose',
                                                   expected)
        # unpack the arguments used for the previous assertion call
        call_args, call_kwargs = self.pipe.producer.send.call_args
        crawl_args_dict = ujson.loads(call_args[1])
        decoded_string = base64.b64decode(crawl_args_dict['body']).decode(crawl_args_dict['encoding'])
        self.assertEquals(decoded_string, item.get('body').decode(item.get('encoding')))
        # Test again against the original (before it was encoded in iso) string
        self.assertEquals(decoded_string, u"αυτό είναι ένα τεστ\n")

        # test kafka exception
        item = self._get_item()
        copy = deepcopy(item)
        copy['success'] = False
        copy['exception'] = 'traceback'

        # send should not crash the pipeline
        self.pipe.producer.send = MagicMock(side_effect=KafkaTimeoutError('bad kafka'))
        ret_val = self.pipe.process_item(item, spider)




'''
Offline tests
'''
import mock
from mock import MagicMock
from unittest import TestCase
from crawling.log_retry_middleware import LogRetryMiddleware


class TestLogRetryMiddlewareStats(TestCase):

    @mock.patch('crawling.log_retry_middleware.LogRetryMiddleware' \
                '.setup')
    def setUp(self, s):
        self.lrm = LogRetryMiddleware(MagicMock())
        self.lrm.settings = {}
        self.lrm.name = 'OverrideSpider'
        self.lrm.redis_conn = MagicMock()
        self.lrm.logger = MagicMock()
        self.lrm.settings['STATS_CYCLE'] = 5
        self.lrm.settings['STATS_TIMES'] = []
        self.lrm._get_hostname = MagicMock(return_value='host1')

    def test_lrm_stats_setup(self):
        self.lrm.stats_dict = {}

        # test nothing
        self.lrm._setup_stats_status_codes()
        self.assertEqual([str(x) for x in self.lrm.stats_dict.keys()], ['lifetime'])

        # test good/bad rolling stats
        self.lrm.stats_dict = {}
        self.lrm.settings['STATS_TIMES'] = [
            'SECONDS_15_MINUTE',
            'SECONDS_1_HOUR',
            'SECONDS_DUMB',
        ]
        good = [
            'lifetime',  # for totals, not DUMB
            '900',
            '3600',
        ]

        # check that both keys are set up
        self.lrm._setup_stats_status_codes()
        self.assertEqual(
            sorted([str(x) for x in self.lrm.stats_dict.keys()]),
            sorted(good))

        k1 = 'stats:crawler:host1:OverrideSpider:504'

        for time_key in self.lrm.stats_dict:
            if time_key == 0:
                self.assertEqual(
                    self.lrm.stats_dict[0].key,
                    '{k}:lifetime'.format(k=k1)
                    )
            else:
                self.assertEqual(
                    self.lrm.stats_dict[time_key].key,
                    '{k}:{t}'.format(k=k1, t=time_key)
                    )



from __future__ import print_function
import sys


def main():
    '''
    A manual configuration file pusher for the crawlers. This will update
    Zookeeper with the contents of the file specified in the args.
    '''
    import argparse
    from kazoo.client import KazooClient

    parser = argparse.ArgumentParser(
            description="Crawler config file pusher to Zookeeper")
    parser.add_argument('-f', '--file', action='store', required=True,
                        help="The yaml file to use")
    parser.add_argument('-i', '--id', action='store', default="all",
                        help="The crawler id to use in zookeeper")
    parser.add_argument('-p', '--path', action='store',
                        default="/scrapy-cluster/crawler/",
                        help="The zookeeper path to use")
    parser.add_argument('-w', '--wipe', action='store_const', const=True,
                        help="Remove the current config")
    parser.add_argument('-z', '--zoo-keeper', action='store', required=True,
                        help="The Zookeeper connection <host>:<port>")
    args = vars(parser.parse_args())

    filename = args['file']
    id = args['id']
    wipe = args['wipe']
    zoo = args['zoo_keeper']
    path = args['path']

    zk = KazooClient(hosts=zoo)
    zk.start()

    # ensure path exists
    zk.ensure_path(path)

    bytes = open(filename, 'rb').read()

    if zk.exists(path):
        # push the conf file
        if not zk.exists(path + id) and not wipe:
            print("creaing conf node")
            zk.create(path + id, bytes)
        elif not wipe:
            print("updating conf file")
            zk.set(path + id, bytes)

        if wipe:
            zk.set(path + id, None)

    zk.stop()

if __name__ == "__main__":
    sys.exit(main())



'''
Offline tests
'''
from mock import MagicMock
from unittest import TestCase
from scrapy.http import Request
from crawling.redis_dupefilter import RFPDupeFilter


class TestRedisDupefilter(TestCase):

    def setUp(self):
        self.dupe = RFPDupeFilter(MagicMock(), 'key', 1)

    def test_dupe_filter(self):
        req = Request('http://example.com')
        req.meta['crawlid'] = "abc123"

        self.dupe.server.expire = MagicMock()

        # successfully added
        self.dupe.server.sadd = MagicMock(return_value=1)
        self.assertFalse(self.dupe.request_seen(req))

        # unsuccessfully added
        self.dupe.server.sadd = MagicMock(return_value=0)
        self.assertTrue(self.dupe.request_seen(req))



from bhfutils.scutils.method_timer import MethodTimer
from time import sleep

def timeout(sleep_time, wait_time, default, mul_value):
    # define a hidden method to sleep and wait based on parameters
    @MethodTimer.timeout(wait_time, default)
    def _hidden(m_value):
        sleep(sleep_time)
        return m_value * m_value
    # call the newly declared function
    return _hidden(mul_value)

print(timeout(5, 3, "did not finish 2*2", 2))
print(timeout(3, 5, "did not finish 3*3", 3))
print(timeout(2, 1, "did not finish 4*4", 4))


import sys
from os import path
from scrapy.http import Request

from bhfutils.scutils.log_factory import LogFactory


class MetaPassthroughMiddleware(object):

    def __init__(self, settings):
        self.setup(settings)

    def setup(self, settings):
        '''
        Does the actual setup of the middleware
        '''
        # set up the default sc logger
        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SC_LOGGER_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)

        self.logger = LogFactory.get_instance(json=my_json,
                                         name=my_name,
                                         stdout=my_output,
                                         level=my_level,
                                         dir=my_dir,
                                         file=my_file,
                                         bytes=my_bytes,
                                         backups=my_backups)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def process_spider_output(self, response, result, spider):
        '''
        Ensures the meta data from the response is passed
        through in any Request's generated from the spider
        '''
        self.logger.debug("processing meta passthrough middleware")
        for x in result:
            # only operate on requests
            if isinstance(x, Request):
                self.logger.debug("found request")
                # pass along all known meta fields, only if
                # they were not already set in the spider's new request
                for key in list(response.meta.keys()):
                    if key not in x.meta and key != 'playwright_page_methods':
                        x.meta[key] = response.meta[key]
            yield x



from __future__ import absolute_import

import re
import sys
import time
import yaml
import uuid
import ujson
import redis
import random
import socket
import tldextract
from builtins import str
from . import bhf_signals
from builtins import object
from six import string_types
from scrapy.http import Request
from future import standard_library

standard_library.install_aliases()
import urllib.request, urllib.error, urllib.parse
from scrapy.utils.project import get_project_settings
from scrapy.utils.reqser import request_to_dict, request_from_dict

from .redis_dupefilter import RFPDupeFilter
from kazoo.handlers.threading import KazooTimeoutError
from .redis_domain_max_page_filter import RFDomainMaxPageFilter
from .redis_global_page_per_domain_filter import RFGlobalPagePerDomainFilter

from bhfutils.scutils.log_factory import LogFactory
from bhfutils.scutils.redis_queue import RedisPriorityQueue
from bhfutils.scutils.zookeeper_watcher import ZookeeperWatcher
from bhfutils.scutils.redis_throttled_queue import RedisThrottledQueue

settings = get_project_settings()


class DistributedScheduler(object):
    '''
    Scrapy request scheduler that utilizes Redis Throttled Priority Queues
    to moderate different domain scrape requests within a distributed scrapy
    cluster
    '''
    redis_conn = None  # the redis connection
    queue_dict = None  # the dict of throttled queues
    spider = None  # the spider using this scheduler
    queue_keys = None  # the list of current queues
    queue_class = None  # the class to use for the queue
    dupefilter = None  # the redis dupefilter
    global_page_per_domain_filter = None  # the global redis page per domain filter, applied to all domains.
    domain_max_page_filter = None  # the individual domain's redis max page filter.
    update_time = 0  # the last time the queues were updated
    update_ip_time = 0  # the last time the ip was updated
    update_interval = 0  # how often to update the queues
    extract = None  # the tld extractor
    hits = 0  # default number of hits for a queue
    window = 0  # default window to calculate number of hits
    my_ip = None  # the ip address of the scheduler (if needed)
    old_ip = None  # the old ip for logging
    signals = None  # scrapy signals
    ip_update_interval = 0  # the interval to update the ip address
    ip_request_proxy = None
    add_type = None  # add spider type to redis throttle queue key
    add_ip = None  # add spider public ip to redis throttle queue key
    item_retries = 0  # the number of extra tries to get an item
    my_uuid = None  # the generated UUID for the particular scrapy process
    # Zookeeper Dynamic Config Vars
    domain_config = {}  # The list of domains and their configs
    my_id = None  # The id used to read the throttle config
    config_flag = False  # Flag to reload queues if settings are wiped too
    assign_path = None  # The base assigned configuration path to read
    zoo_client = None  # The KazooClient to manage the config
    my_assignment = None  # Zookeeper path to read actual yml config
    black_domains = []  # the domains to ignore thanks to zookeeper config

    def __init__(self, server, persist, update_int, timeout, retries, logger,
                 signals, hits, window, mod, ip_refresh, ip_proxy, add_type, add_ip,
                 ip_regex, backlog_blacklist, queue_timeout, global_page_per_domain_limit,
                 global_page_per_domain_limit_timeout, domain_max_page_timeout):
        '''
        Initialize the scheduler
        '''
        self.redis_conn = server
        self.persist = persist
        self.queue_dict = {}
        self.update_interval = update_int
        self.hits = hits
        self.window = window
        self.moderated = mod
        self.rfp_timeout = timeout
        self.ip_update_interval = ip_refresh
        self.ip_request_proxy = ip_proxy
        self.add_type = add_type
        self.add_ip = add_ip
        self.item_retries = retries
        self.logger = logger
        self.ip_regex = re.compile(ip_regex)
        self.backlog_blacklist = backlog_blacklist
        self.queue_timeout = queue_timeout
        self.global_page_per_domain_limit = global_page_per_domain_limit
        self.global_page_per_domain_limit_timeout = global_page_per_domain_limit_timeout
        self.domain_max_page_timeout = domain_max_page_timeout

        # set up tldextract
        self.extract = tldextract.TLDExtract()

        self.signals = signals
        self.update_ip_address()

        # if we need better uuid's mod this line
        self.my_uuid = str(uuid.uuid4()).split('-')[4]

    def setup_zookeeper(self):
        self.assign_path = settings.get('ZOOKEEPER_ASSIGN_PATH', "")
        self.my_id = settings.get('ZOOKEEPER_ID', 'all')
        self.logger.debug("Trying to establish Zookeeper connection")
        try:
            self.zoo_watcher = ZookeeperWatcher(
                hosts=settings.get('ZOOKEEPER_HOSTS'),
                filepath=self.assign_path + self.my_id,
                config_handler=self.change_config,
                error_handler=self.error_config,
                pointer=False, ensure=True, valid_init=True)
        except KazooTimeoutError:
            self.logger.error("Could not connect to Zookeeper")
            sys.exit(1)

        if self.zoo_watcher.ping():
            self.logger.debug("Successfully set up Zookeeper connection")
        else:
            self.logger.error("Could not ping Zookeeper")
            # sys.exit(1)

    def change_config(self, config_string):
        if config_string and len(config_string) > 0:
            loaded_config = yaml.safe_load(config_string)
            self.logger.info("Zookeeper config changed", extra=loaded_config)
            self.load_domain_config(loaded_config)
            self.update_domain_queues()
        elif config_string is None or len(config_string) == 0:
            self.error_config("Zookeeper config wiped")

        self.create_queues()

    def load_domain_config(self, loaded_config):
        '''
        Loads the domain_config and sets up queue_dict
        @param loaded_config: the yaml loaded config dict from zookeeper
        '''
        self.domain_config = {}
        # vetting process to ensure correct configs
        if loaded_config:
            if 'domains' in loaded_config:
                for domain in loaded_config['domains']:
                    item = loaded_config['domains'][domain]
                    # check valid
                    if 'window' in item and 'hits' in item:
                        self.logger.debug("Added domain {dom} to loaded config"
                                          .format(dom=domain))
                        self.domain_config[domain] = item
            if 'blacklist' in loaded_config:
                self.black_domains = loaded_config['blacklist']

        self.config_flag = True

    def update_domain_queues(self):
        '''
        Check to update existing queues already in memory
        new queues are created elsewhere
        '''
        for key in self.domain_config:
            final_key = "{name}:{domain}:queue".format(
                name=self.spider.name,
                domain=key)
            # we already have a throttled queue for this domain, update it to new settings
            if final_key in self.queue_dict:
                self.queue_dict[final_key][0].window = float(self.domain_config[key]['window'])
                self.logger.debug("Updated queue {q} with new config"
                                  .format(q=final_key))
                # if scale is applied, scale back; otherwise use updated hits
                if 'scale' in self.domain_config[key]:
                    # round to int
                    hits = int(self.domain_config[key]['hits'] * self.fit_scale(
                        self.domain_config[key]['scale']))
                    self.queue_dict[final_key][0].limit = float(hits)
                else:
                    self.queue_dict[final_key][0].limit = float(self.domain_config[key]['hits'])

    def error_config(self, message):
        extras = {}
        extras['message'] = message
        extras['revert_window'] = self.window
        extras['revert_hits'] = self.hits
        extras['spiderid'] = self.spider.name
        self.logger.info("Lost config from Zookeeper", extra=extras)
        # lost connection to zookeeper, reverting back to defaults
        for key in self.domain_config:
            final_key = "{name}:{domain}:queue".format(
                name=self.spider.name,
                domain=key)
            self.queue_dict[final_key][0].window = self.window
            self.queue_dict[final_key][0].limit = self.hits

        self.domain_config = {}

    def fit_scale(self, scale):
        '''
        @return: a scale >= 0 and <= 1
        '''
        if scale >= 1:
            return 1.0
        elif scale <= 0:
            return 0.0
        else:
            return scale

    def create_queues(self):
        '''
        Updates the in memory list of the redis queues
        Creates new throttled queue instances if it does not have them
        '''
        # new config could have loaded between scrapes
        newConf = self.check_config()

        self.queue_keys = self.redis_conn.keys(self.spider.name + ":*:queue")

        for key in self.queue_keys:
            # build final queue key, depending on type and ip bools
            throttle_key = ""

            if self.add_type:
                throttle_key = self.spider.name + ":"
            if self.add_ip:
                throttle_key = throttle_key + self.my_ip + ":"

            # add the tld from the key `type:tld:queue`
            the_domain = re.split(':', key)[1]
            throttle_key = throttle_key + the_domain

            if key not in self.queue_dict or newConf:
                self.logger.debug("Added new Throttled Queue {q}"
                                  .format(q=key))
                q = RedisPriorityQueue(self.redis_conn, key, encoding=ujson)

                # use default window and hits
                if the_domain not in self.domain_config:
                    # this is now a tuple, all access needs to use [0] to get
                    # the object, use [1] to get the time
                    self.queue_dict[key] = [RedisThrottledQueue(self.redis_conn,
                                                                q, self.window, self.hits, self.moderated, throttle_key,
                                                                throttle_key, True), time.time()]
                # use custom window and hits
                else:
                    window = self.domain_config[the_domain]['window']
                    hits = self.domain_config[the_domain]['hits']

                    # adjust the crawl rate based on the scale if exists
                    if 'scale' in self.domain_config[the_domain]:
                        hits = int(hits * self.fit_scale(self.domain_config[the_domain]['scale']))

                    self.queue_dict[key] = [RedisThrottledQueue(self.redis_conn,
                                                                q, window, hits, self.moderated, throttle_key,
                                                                throttle_key, True), time.time()]

    def expire_queues(self):
        '''
        Expires old queue_dict keys that have not been used in a long time.
        Prevents slow memory build up when crawling lots of different domains
        '''
        curr_time = time.time()
        for key in list(self.queue_dict):
            diff = curr_time - self.queue_dict[key][1]
            if diff > self.queue_timeout:
                self.logger.debug("Expiring domain queue key " + key)
                del self.queue_dict[key]
                if key in self.queue_keys:
                    self.queue_keys.remove(key)

    def check_config(self):
        '''
        Controls configuration for the scheduler
        @return: True if there is a new configuration
        '''
        if self.config_flag:
            self.config_flag = False
            return True

        return False

    def update_ip_address(self):
        '''
        Updates the scheduler so it knows its own ip address
        '''
        # assign local ip in case of exception
        self.old_ip = self.my_ip
        self.my_ip = '127.0.0.1'
        try:
            req = urllib.request.Request(url=settings.get('PUBLIC_IP_URL',
                                                          'http://ip.42.pl/raw'))
            if self.ip_request_proxy is not None:
                req.set_proxy(self.ip_request_proxy, 'http')
            obj = urllib.request.urlopen(req)
            results = self.ip_regex.findall(obj.read().decode('utf-8'))
            if len(results) > 0:
                self.my_ip = results[0]
            else:
                raise IOError("Could not get valid IP Address")
            obj.close()
            self.logger.debug("Current public ip: {ip}".format(ip=self.my_ip))

            if self.old_ip != self.my_ip:
                self.logger.info("Changed Public IP: {old} -> {new}".format(
                    old=self.old_ip, new=self.my_ip))
                if self.old_ip is not None and self.my_ip != '52.28.30.162':
                    self.signals.send_catch_log_deferred(signal=bhf_signals.ip_changed)
        except IOError:
            self.logger.error("Could not reach out to get public ip")
            pass

    def report_self(self):
        '''
        Reports the crawler uuid to redis
        '''
        self.logger.debug("Reporting self id", extra={'uuid': self.my_uuid})
        key = "stats:crawler:{m}:{s}:{u}".format(
            m=socket.gethostname(),
            s=self.spider.name,
            u=self.my_uuid)
        self.redis_conn.set(key, time.time())
        self.redis_conn.expire(key, self.ip_update_interval * 2)

    @classmethod
    def from_settings(cls, signals, settings):
        server = redis.Redis(host=settings.get('REDIS_HOST'),
                             port=settings.get('REDIS_PORT'),
                             db=settings.get('REDIS_DB'),
                             password=settings.get('REDIS_PASSWORD'),
                             decode_responses=True,
                             socket_timeout=settings.get('REDIS_SOCKET_TIMEOUT'),
                             socket_connect_timeout=settings.get('REDIS_SOCKET_TIMEOUT'))
        persist = settings.get('SCHEDULER_PERSIST', True)
        up_int = settings.get('SCHEDULER_QUEUE_REFRESH', 10)
        hits = settings.get('QUEUE_HITS', 10)
        window = settings.get('QUEUE_WINDOW', 60)
        mod = settings.get('QUEUE_MODERATED', False)
        timeout = settings.get('DUPEFILTER_TIMEOUT', 600)
        ip_refresh = settings.get('SCHEDULER_IP_REFRESH', 60)
        ip_proxy = settings.get('SCHEDULER_IP_PROXY')
        add_type = settings.get('SCHEDULER_TYPE_ENABLED', False)
        add_ip = settings.get('SCHEDULER_IP_ENABLED', False)
        retries = settings.get('SCHEUDLER_ITEM_RETRIES', 3)
        ip_regex = settings.get('IP_ADDR_REGEX', '.*')
        backlog_blacklist = settings.get('SCHEDULER_BACKLOG_BLACKLIST', True)
        queue_timeout = settings.get('SCHEDULER_QUEUE_TIMEOUT', 3600)

        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SC_LOGGER_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)

        logger = LogFactory.get_instance(json=my_json,
                                         name=my_name,
                                         stdout=my_output,
                                         level=my_level,
                                         dir=my_dir,
                                         file=my_file,
                                         bytes=my_bytes,
                                         backups=my_backups)

        global_page_per_domain_limit = settings.get('GLOBAL_PAGE_PER_DOMAIN_LIMIT', None)
        global_page_per_domain_limit_timeout = settings.get('GLOBAL_PAGE_PER_DOMAIN_LIMIT_TIMEOUT', 600)
        domain_max_page_timeout = settings.get('DOMAIN_MAX_PAGE_TIMEOUT', 600)

        return cls(server, persist, up_int, timeout, retries, logger, signals,
                   hits, window, mod, ip_refresh, ip_proxy, add_type, add_ip, ip_regex,
                   backlog_blacklist, queue_timeout, global_page_per_domain_limit,
                   global_page_per_domain_limit_timeout, domain_max_page_timeout)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.signals, crawler.settings)

    def open(self, spider):
        self.spider = spider
        self.spider.set_logger(self.logger)
        self.create_queues()
        self.setup_zookeeper()
        self.dupefilter = RFPDupeFilter(self.redis_conn,
                                        self.spider.name + ':dupefilter',
                                        self.rfp_timeout)
        self.global_page_per_domain_filter = RFGlobalPagePerDomainFilter(self.redis_conn,
                                                                         self.spider.name + ':global_page_count_filter',
                                                                         self.global_page_per_domain_limit,
                                                                         self.global_page_per_domain_limit_timeout)
        self.domain_max_page_filter = RFDomainMaxPageFilter(self.redis_conn,
                                                            self.spider.name + ':domain_max_page_filter',
                                                            self.domain_max_page_timeout)

    def close(self, reason):
        self.logger.info("Closing Spider", {'spiderid': self.spider.name})
        if not self.persist:
            self.logger.warning("Clearing crawl queues")
            self.dupefilter.clear()
            self.global_page_per_domain_filter.clear()
            self.domain_max_page_filter.clear()
            for key in self.queue_keys:
                self.queue_dict[key][0].clear()

    def is_blacklisted(self, appid, crawlid):
        '''
        Checks the redis blacklist for crawls that should not be propagated
        either from expiring or stopped
        @return: True if the appid crawlid combo is blacklisted
        '''
        key_check = '{appid}||{crawlid}'.format(appid=appid,
                                                crawlid=crawlid)
        redis_key = self.spider.name + ":blacklist"
        return self.redis_conn.sismember(redis_key, key_check)

    def enqueue_request(self, request):
        '''
        Pushes a request from the spider into the proper throttled queue
        '''

        # # # # # # # # # # # # # # # # # # Duplicate link Filter # # # # # # # # # # # # # # #
        if not request.dont_filter and self.dupefilter.request_seen(request):
            self.logger.debug("Request not added back to redis")
            return

        # An individual crawling request of a domain's page
        req_dict = request_to_dict(request, self.spider)

        # # # # # # # # # # # # # # # # # # Page Limit Filters # # # # # # # # # # # # # # #
        # Max page filter per individual domain
        if req_dict['meta']['domain_max_pages'] and self.domain_max_page_filter.request_page_limit_reached(
                request=request,
                spider=self.spider):
            self.logger.debug("Request {0} reached domain's page limit of {1}".format(
                request.url,
                req_dict['meta']['domain_max_pages']))
            return

        # Global - cluster wide - max page filter
        if self.global_page_per_domain_limit and self.global_page_per_domain_filter.request_page_limit_reached(
                request=request,
                spider=self.spider):
            self.logger.debug("Request {0} reached global page limit of {1}".format(
                request.url,
                self.global_page_per_domain_limit))
            return

        # # # # # # # # # # # # # # # # # # Blacklist Filter # # # # # # # # # # # # # # #
        if not self.is_blacklisted(req_dict['meta']['appid'],
                                   req_dict['meta']['crawlid']):
            # grab the tld of the request
            ex_res = self.extract(req_dict['url'])
            key = "{sid}:{dom}.{suf}:queue".format(
                sid=req_dict['meta']['spiderid'],
                dom=ex_res.domain,
                suf=ex_res.suffix)

            curr_time = time.time()

            domain = "{d}.{s}".format(d=ex_res.domain, s=ex_res.suffix)

            # allow only if we want all requests or we want
            # everything but blacklisted domains
            # insert if crawl never expires (0) or time < expires
            if (self.backlog_blacklist or
                (not self.backlog_blacklist and
                 domain not in self.black_domains)) and \
                    (req_dict['meta']['expires'] == 0 or
                     curr_time < req_dict['meta']['expires']):
                # we may already have the queue in memory
                if key in self.queue_keys:
                    self.queue_dict[key][0].push(req_dict,
                                                 req_dict['meta']['priority'])
                else:
                    # shoving into a new redis queue, negative b/c of sorted sets
                    # this will populate ourself and other schedulers when
                    # they call create_queues
                    self.redis_conn.zadd(key,
                                         {ujson.dumps(req_dict, reject_bytes=False): -req_dict['meta']['priority']})
                self.logger.debug("Crawlid: '{id}' Appid: '{appid}' added to queue"
                                  .format(appid=req_dict['meta']['appid'],
                                          id=req_dict['meta']['crawlid']))
            else:
                self.logger.debug("Crawlid: '{id}' Appid: '{appid}' expired"
                                  .format(appid=req_dict['meta']['appid'],
                                          id=req_dict['meta']['crawlid']))
        else:
            self.logger.debug("Crawlid: '{id}' Appid: '{appid}' blacklisted"
                              .format(appid=req_dict['meta']['appid'],
                                      id=req_dict['meta']['crawlid']))

    def find_item(self):
        '''
        Finds an item from the throttled queues
        '''
        random.shuffle(self.queue_keys)
        count = 0

        while count <= self.item_retries:
            for key in self.queue_keys:
                # skip if the whole domain has been blacklisted in zookeeper
                if key.split(':')[1] in self.black_domains:
                    continue
                # the throttled queue only returns an item if it is allowed
                item = self.queue_dict[key][0].pop()

                if item:
                    # update timeout and return
                    self.queue_dict[key][1] = time.time()
                    return item

            count = count + 1

        return None

    def next_request(self):
        '''
        Logic to handle getting a new url request, from a bunch of
        different queues
        '''
        t = time.time()
        # update the redis queues every so often
        if t - self.update_time > self.update_interval:
            self.update_time = t
            self.create_queues()
            self.expire_queues()

        # update the ip address every so often
        if t - self.update_ip_time > self.ip_update_interval:
            self.update_ip_time = t
            self.update_ip_address()
            self.report_self()

        item = self.find_item()
        if item:
            self.logger.debug(u"Found url to crawl {url}" \
                              .format(url=item['url']))
            if 'meta' in item:
                # item is a serialized request
                req = request_from_dict(item, self.spider)
            else:
                # item is a feed from outside, parse it manually
                req = self.request_from_feed(item)

            # extra check to add items to request
            if 'useragent' in req.meta and req.meta['useragent'] is not None:
                req.headers['User-Agent'] = req.meta['useragent']
            if 'cookie' in req.meta and req.meta['cookie'] is not None:
                if isinstance(req.meta['cookie'], dict):
                    req.cookies = req.meta['cookie']
                elif isinstance(req.meta['cookie'], string_types):
                    req.cookies = self.parse_cookie(req.meta['cookie'])

            return req

        return None

    def request_from_feed(self, item):
        try:
            req = Request(item['url'])
        except ValueError:
            # need absolute url
            # need better url validation here
            req = Request('http://' + item['url'])

        # defaults not in schema
        if 'curdepth' not in item:
            item['curdepth'] = 0
        if "retry_times" not in item:
            item['retry_times'] = 0

        for key in list(item.keys()):
            req.meta[key] = item[key]

        # extra check to add items to request
        if 'cookie' in item and item['cookie'] is not None:
            if isinstance(item['cookie'], dict):
                req.cookies = item['cookie']
            elif isinstance(item['cookie'], string_types):
                req.cookies = self.parse_cookie(item['cookie'])
        return req

    def parse_cookie(self, string):
        '''
        Parses a cookie string like returned in a Set-Cookie header
        @param string: The cookie string
        @return: the cookie dict
        '''
        results = re.findall('([^=]+)=([^\;]+);?\s?', string)
        my_dict = {}
        for item in results:
            my_dict[item[0]] = item[1]

        return my_dict

    def has_pending_requests(self):
        '''
        We never want to say we have pending requests
        If this returns True scrapy sometimes hangs.
        '''
        return False



import sys
import json
import copy
import ujson
import pickle
from os import path
from builtins import object
from unittest import TestCase

from mock import MagicMock

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

from scutils.redis_queue import Base, RedisQueue, RedisPriorityQueue, RedisStack


class TestBase(TestCase):

    def setUp(self):
        self.q = Base(MagicMock(), 'k')

    def test_init(self):
        # assert example good encoding
        q = Base(MagicMock(), 'key')
        q1 = Base(MagicMock(), 'key', pickle)
        q2 = Base(MagicMock(), 'key', ujson)
        q3 = Base(MagicMock(), 'key', json)

        # assert bad encodings
        with self.assertRaises(NotImplementedError) as e:
            q2 = Base(MagicMock(), 'key', copy)

    def test_encode(self):
        q = Base(MagicMock(), 'key', pickle)
        # python pickling is different between versions
        data = pickle.dumps('cool', protocol=-1).decode('latin1')
        self.assertEqual(q._encode_item('cool'), data)
        q2 = Base(MagicMock(), 'key', ujson)
        self.assertEqual(q2._encode_item('cool2'), '"cool2"')

    def test_decode(self):
        q = Base(MagicMock(), 'key', pickle)
        self.assertEqual(q._decode_item(u"\x80\x02U\x04coolq\x00."), 'cool')

        q2 = Base(MagicMock(), 'key', ujson)
        self.assertEqual(q2._decode_item('"cool2"'), 'cool2')

    def test_len(self):
        with self.assertRaises(NotImplementedError):
            len(self.q)

    def test_push(self):
        with self.assertRaises(NotImplementedError):
            self.q.push("blah")

    def test_pop(self):
        with self.assertRaises(NotImplementedError):
            self.q.pop()

    def test_clear(self):
        self.q.clear()


class QueueMixin(object):
    queue_class = None

    def setUp(self):
        self.queue = self.queue_class(MagicMock(), 'key', ujson)


class TestRedisQueue(QueueMixin, TestCase):

    queue_class = RedisQueue

    def test_len(self):
        self.assertTrue(hasattr(self.queue, '__len__'))
        len(self.queue)

    def test_push(self):
        self.assertTrue(hasattr(self.queue, 'push'))
        self.queue.push("blah")

    def test_pop(self):
        self.assertTrue(hasattr(self.queue, 'pop'))
        self.queue.server.rpop = MagicMock(return_value='"stuff"')
        self.assertEqual(self.queue.pop(), "stuff")


class TestRedisPriorityQueue(QueueMixin, TestCase):

    queue_class = RedisPriorityQueue

    def test_len(self):
        self.assertTrue(hasattr(self.queue, '__len__'))
        len(self.queue)

    def test_push(self):
        self.assertTrue(hasattr(self.queue, 'push'))
        self.queue.push("blah", 60)

    def test_pop(self):
        self.assertTrue(hasattr(self.queue, 'pop'))
        m = MagicMock()
        m.execute = MagicMock(return_value=[{}, 60])
        self.queue.server.pipeline = MagicMock(return_value=m)
        self.assertEqual(self.queue.pop(), None)


class TestRedisStack(QueueMixin, TestCase):

    queue_class = RedisStack

    def test_len(self):
        self.assertTrue(hasattr(self.queue, '__len__'))
        len(self.queue)

    def test_push(self):
        self.assertTrue(hasattr(self.queue, 'push'))
        self.queue.push("blah")

    def test_pop(self):
        self.assertTrue(hasattr(self.queue, 'pop'))
        self.queue.server.lpop = MagicMock(return_value='"stuff"')
        self.assertEqual(self.queue.pop(), "stuff")



import sys


def main():

    import argparse
    import redis
    import time
    import random

    import sys
    from os import path
    sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))

    from bhfutils.scutils.redis_queue import RedisPriorityQueue
    from bhfutils.scutils.redis_throttled_queue import RedisThrottledQueue

    parser = argparse.ArgumentParser(description="Throttled Queue Test Script."
                    " Start either a single or multiple processes to see the "
                " throttled queue mechanism in action.")
    parser.add_argument('-r', '--redis-host', action='store', required=True,
                        help="The Redis host ip")
    parser.add_argument('-p', '--redis-port', action='store', default='6379',
                        help="The Redis port")
    parser.add_argument('-P', '--redis-password', action='store', default=None,
                        help="The Redis password")
    parser.add_argument('-m', '--moderate', action='store_const', const=True,
                        default=False, help="Moderate the outbound Queue")
    parser.add_argument('-w', '--window', action='store', default=60,
                        help="The window time to test")
    parser.add_argument('-n', '--num-hits', action='store', default=10,
                        help="The number of pops allowed in the given window")
    parser.add_argument('-q', '--queue', action='store', default='testqueue',
                        help="The Redis queue name")
    parser.add_argument('-e', '--elastic', action='store_const', const=True,
                        default=False, help="Test variable elastic catch up"
                        " with moderation")

    args = vars(parser.parse_args())

    window = int(args['window'])
    num = int(args['num_hits'])
    host = args['redis_host']
    port = args['redis_port']
    password = args['redis_password']
    mod = args['moderate']
    queue = args['queue']
    elastic = args['elastic']

    conn = redis.Redis(host=host, port=port, password=password, decode_responses=True)

    q = RedisPriorityQueue(conn, queue)
    t = RedisThrottledQueue(conn, q, window, num, mod, elastic=elastic)

    def push_items(amount):
        for i in range(0, amount):
            t.push('item-'+str(i), i)

    print("Adding", num * 2, "items for testing")
    push_items(num * 2)

    def read_items():
        print("Kill when satisfied ^C")
        ti = time.time()
        count = 0
        while True:
            item = t.pop()
            if item:
                print("My item", item, "My time:", time.time() - ti)
                count += 1

            if elastic:
                time.sleep(int(random.random() * (t.moderation * 3)))

    try:
        read_items()
    except KeyboardInterrupt:
        pass
    t.clear()
    print("Finished")

if __name__ == "__main__":
    sys.exit(main())



import sys
from setuptools import setup, find_packages


def readme():
    ''' Returns README.rst contents as str '''
    with open('README.rst') as f:
        return f.read()


install_requires = [
    'python-json-logger==0.1.8',
    'ConcurrentLogHandler>=0.9.1',
    'redis>=4.0.2',
    'kazoo>=2.8.0',
    'mock>=4.0.3',
    'playwright>=1.20.1',
    'testfixtures>=6.18.3',
    'ujson>=4.3.0',
    'future>=0.18.2'
]

lint_requires = [
    'pep8',
    'pyflakes'
]

tests_require = [
    'mock>=2.0.0',
    'testfixtures>=4.13.5'
]

dependency_links = []
setup_requires = []
extras_require = {
    'test': tests_require,
    'all': install_requires + tests_require,
    'docs': ['sphinx'] + tests_require,
    'lint': lint_requires
}

if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

setup(
    name='bhfutils',
    version='0.0.42',
    description='Utilities that are used by any spider of Behoof project',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    author='Teplygin Vladimir',
    author_email='vvteplygin@gmail.com',
    license='MIT',
    url='https://behoof.app/',
    keywords=['behoof', 'scrapy-cluster', 'utilities'],
    packages=find_packages(),
    package_data={},
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require=extras_require,
    dependency_links=dependency_links,
    zip_safe=True,
    include_package_data=True,
)



from builtins import object
# -*- coding: utf-8 -*-

# Define your item pipelines here

import sys
import ujson
import base64
import traceback
from os import path
import datetime as dt

from kafka import KafkaProducer
from kafka.errors import KafkaTimeoutError
from bhfutils.crawler.items import RawResponseItem

sys.path.append(path.dirname(path.dirname(path.abspath('utils'))))

from bhfutils.scutils.log_factory import LogFactory


class LoggingBeforePipeline(object):

    '''
    Logs the crawl, currently the 1st priority of the pipeline
    '''

    def __init__(self, logger):
        self.logger = logger
        self.logger.debug("Setup before pipeline")

    @classmethod
    def from_settings(cls, settings):
        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SC_LOGGER_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)

        logger = LogFactory.get_instance(json=my_json,
                                         name=my_name,
                                         stdout=my_output,
                                         level=my_level,
                                         dir=my_dir,
                                         file=my_file,
                                         bytes=my_bytes,
                                         backups=my_backups)

        return cls(logger)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings)

    def process_item(self, item, spider):
        self.logger.debug("Processing item in LoggingBeforePipeline")
        if isinstance(item, RawResponseItem):
            # make duplicate item, but remove unneeded keys
            item_copy = {}
            item_copy['url'] = item['url']
            item_copy['appid'] = item['appid']
            item_copy['spiderid'] = spider.name
            item_copy['crawlid'] = item['crawlid']
            item_copy['statusCode'] = item['statusCode']
            item_copy['responseUrl'] = item['responseUrl']
            item_copy['playgroundId'] = item['playgroundId']
            item_copy['logger'] = self.logger.name
            item_copy['action'] = 'emit'
            self.logger.info('Scraped page', extra=item_copy)
            return item
        else:
            self.logger.warn('Received unknown item')
            return None


class KafkaPipeline(object):
    '''
    Pushes a serialized item to appropriate Kafka topics.
    '''

    def __init__(self, producer, topic_prefix, logger, appids,
                 use_base64):
        self.producer = producer
        self.topic_prefix = topic_prefix
        self.topic_list = []
        self.appid_topics = appids
        self.logger = logger
        self.logger.debug("Setup kafka pipeline")
        self.use_base64 = use_base64

    @classmethod
    def from_settings(cls, settings):
        my_level = settings.get('SC_LOG_LEVEL', 'INFO')
        my_name = settings.get('SC_LOGGER_NAME', 'sc-logger')
        my_output = settings.get('SC_LOG_STDOUT', True)
        my_json = settings.get('SC_LOG_JSON', False)
        my_dir = settings.get('SC_LOG_DIR', 'logs')
        my_bytes = settings.get('SC_LOG_MAX_BYTES', '10MB')
        my_file = settings.get('SC_LOG_FILE', 'main.log')
        my_backups = settings.get('SC_LOG_BACKUPS', 5)
        my_appids = settings.get('KAFKA_APPID_TOPICS', False)

        logger = LogFactory.get_instance(json=my_json,
                                         name=my_name,
                                         stdout=my_output,
                                         level=my_level,
                                         dir=my_dir,
                                         file=my_file,
                                         bytes=my_bytes,
                                         backups=my_backups)

        try:
            producer = KafkaProducer(bootstrap_servers=settings['KAFKA_HOSTS'],
                                     retries=3,
                                     linger_ms=settings['KAFKA_PRODUCER_BATCH_LINGER_MS'],
                                     buffer_memory=settings['KAFKA_PRODUCER_BUFFER_BYTES'],
                                     value_serializer=lambda m: m.encode('utf-8'),
                                     max_request_size=settings['KAFKA_PRODUCER_MAX_REQUEST_SIZE'])
        except Exception as e:
                logger.error("Unable to connect to Kafka in Pipeline"\
                    ", raising exit flag.")
                # this is critical so we choose to exit.
                # exiting because this is a different thread from the crawlers
                # and we want to ensure we can connect to Kafka when we boot
                sys.exit(1)
        topic_prefix = settings['KAFKA_TOPIC_PREFIX']
        use_base64 = settings['KAFKA_BASE_64_ENCODE']

        return cls(producer, topic_prefix, logger, appids=my_appids,
                   use_base64=use_base64)

    @classmethod
    def from_crawler(cls, crawler):
        return cls.from_settings(crawler.settings)

    def _get_time(self):
        '''
        Returns an ISO formatted string of the current time
        '''
        return dt.datetime.utcnow().isoformat()

    def _clean_item(self, item):
        '''
        Cleans the item to be logged
        '''
        item_copy = dict(item)
        item_copy['action'] = 'ack'
        item_copy['logger'] = self.logger.name

        return item_copy

    def _kafka_success(self, item, spider, response):
        '''
        Callback for successful send
        '''
        item['success'] = True
        item = self._clean_item(item)
        item['spiderid'] = spider.name
        self.logger.info("Sent page to Kafka", item)


    def _kafka_failure(self, item, spider, exception):
        '''
        Callback for failed send
        '''
        item['success'] = False
        item['exception'] = exception if exception else traceback.format_exc()
        item['spiderid'] = spider.name
        item = self._clean_item(item)
        self.logger.error("Failed to send page to Kafka", item)

    def process_item(self, item, spider):
        try:
            self.logger.debug("Processing item in KafkaPipeline")
            datum = dict(item)
            if "url" in datum:
                del datum["url"]
            if "responseUrl" in datum:
                del datum["responseUrl"]
            if "statusCode" in datum:
                del datum["statusCode"]
            datum["timestamp"] = self._get_time()
            prefix = self.topic_prefix

            try:
                # Get the encoding. If it's not a key of datum, return utf-8
                encoding = datum.get('encoding', 'utf-8')
                message = ujson.dumps(datum, sort_keys=True, reject_bytes=False)
            except:
                message = 'json failed to parse'

            firehose_topic = "{prefix}.crawled_firehose".format(prefix=prefix)
            future = self.producer.send(firehose_topic, message)
            future.add_callback(self._kafka_success, datum, spider)
            future.add_errback(self._kafka_failure, datum, spider)

            if self.appid_topics:
                appid_topic = "{prefix}.crawled_{appid}".format(
                        prefix=prefix, appid=datum["appid"])
                future2 = self.producer.send(appid_topic, message)
                future2.add_callback(self._kafka_success, datum, spider)
                future2.add_errback(self._kafka_failure, datum, spider)

        except KafkaTimeoutError:
            self.logger.warning("Caught KafkaTimeoutError exception")

        return item

    def close_spider(self, spider):
        self.logger.info("Closing Kafka Pipeline")
        self.producer.flush()
        self.producer.close(timeout=10)



import warnings

from scrapy.exceptions import ScrapyDeprecationWarning

__all__ = ["PageMethod"]


class PageMethod:
    """
    Represents a method to be called (and awaited if necessary) on a
    Playwright page, such as "click", "screenshot", "evaluate", etc.
    """

    def __init__(self, method: str, *args, **kwargs) -> None:
        self.method = method
        self.args = args
        self.kwargs = kwargs
        self.result = None

    def __str__(self):
        return f"<{self.__class__.__name__} for method '{self.method}'>"

    __repr__ = __str__


class PageCoroutine(PageMethod):
    def __init__(self, method: str, *args, **kwargs) -> None:
        warnings.warn(
            f"The {_qualname(self.__class__)} class is deprecated,"
            f" please use {_qualname(PageMethod)} instead.",
            category=ScrapyDeprecationWarning,
            stacklevel=2,
        )
        super().__init__(method, *args, **kwargs)


def _qualname(cls: type) -> str:
    return f"{cls.__module__}.{cls.__qualname__}"